package chars;

import java.awt.Color;
import java.util.ArrayList;

import maps.Cell;
import maps.Map;
import util.Dir;
import util.Util;
import events.ContactEvent;
import events.Event;

public abstract class Char {

    public static final double RECOVERY_FACTOR = 0.006;
    public static final double CHILD_GENERATION_FACTOR = 0.5;
    public static final double CHILD_DISTANCE_FACTOR = 1.5;
    public static final double GROUP_DENSITY = 0.25;

    protected Cell cell, origCell;
    protected String name;
    protected char symbol;
    protected Color color;
    protected int speed;
    protected int actionSpeed;
    protected double movingPr;
    protected int movingR;
    protected double hp, maxHp;
    protected double mp, maxMp;
    protected boolean hostile;
    protected int sight;
    protected int minDmg, maxDmg;
    protected int familyDegree;
    protected int optChildrenCount;
    protected int level;
    protected Char parent;
    protected ArrayList<Char> children;
    protected Map map;
    protected boolean rideable;
    protected String talk;
    protected boolean hasAffair;

    protected double turnPool = 0.0;

    public Char(String name, Cell origCell) {
	this.name = name;
	this.origCell = origCell;
	this.cell = origCell;
	this.map = cell.getMap();
	this.parent = null;
	this.children = new ArrayList<Char>();

	cell.setChar(this);
    }

    public void doAction() {
    }

    // Target is player
    protected void attack(Char target) {
	int dmg = calcDmg(target);
	target.decreaseHp(dmg);
	String msg = Util.capitalizeFirstLetter(this.name) + " hits you " + dmg + " damage.";
	map.getGameModel().addEvent(new ContactEvent(0, msg, Event.DEFENCE, this)); // TODO: AGAIN: there is something happening with defense, eller hur?
    }

    private int calcDmg(Char target) {
	return Util.randInt(minDmg, maxDmg);
    }

    protected void die() {
	map.getGameModel().getPlayer().addExp(level);
	for (Char child : children) {
	    child.setParent(null);
	}
	if (parent != null)
	    parent.children.remove(this);
	cell.setChar(null);
	map.getChars().remove(this);
    }

    public void setCell(Cell cell) {
	map.getChars().remove(this);
	this.cell = cell;
	this.map = cell.getMap();
	map.getChars().add(this);
    }
    
    public void setOrigCell(Cell cell) {
	origCell = cell;
    }
/*
    public void draw(Graphics2D g) {
	GameModel gameModel = map.getGameModel();
	int drawX = GameView.getDrawX(gameModel, getCell().getX());
	int drawY = GameView.getDrawY(gameModel, getCell().getY());

	g.setColor(color);
	View.drawCharAt(g, symbol, drawX, drawY);
    }*/

    public void move(Cell target) {
	cell.setChar(null);
	cell = target;
	target.setChar(this);
    }
    
    public boolean tryToMove(Dir dir) {		// TODO: no function, just to be overridden
	return true;
    }

    public void decreaseTurnPool(double value) {
	turnPool -= value;
    }

    public void increaseTurnPool(double value) {
	turnPool += value;
    }

    public boolean isNextTo(Char c) {
	return Util.getDistance(cell.getX(), cell.getY(), c.getCell().getX(), c.getCell().getY()) < 1.5;
    }
    
    public double distanceTo(Cell c) {
	return Util.getDistance(cell.getX(), cell.getY(), c.getX(), c.getY());
    }

    public double distanceTo(Char c) {
	return Util.getDistance(cell.getX(), cell.getY(), c.getCell().getX(), c.getCell().getY());
    }

    protected boolean sees(Char c) {
	return Util.getDistance(cell.getX(), cell.getY(), c.getCell().getX(), c.getCell().getY()) < sight;
    }
    
    public Char getAdjacentChar(Dir dir) {
	return cell.getAdjacentCell(dir).getChar();
    }

    public void recovery() {
	increaseHp(maxHp * RECOVERY_FACTOR);
	increaseMp(maxMp * RECOVERY_FACTOR);
    }

    public void increaseHp(double amount) {
	hp += amount;
	if (hp > maxHp)
	    hp = maxHp;
    }

    public void increaseMp(double amount) {
	mp += amount;
	if (mp > maxMp)
	    mp = maxMp;
    }

    public void decreaseHp(double amount) {
	hp -= amount;

	if (hp <= 0) {
	    die();
	}
    }

    public void decreaseMp(double amount) {
	mp -= amount;

	if (mp < 0) {
	    mp = 0;
	}
    }

    public void setParent(Char parent) {
	this.parent = parent;
    }

    public double getGroupRadius() {
	return Math.sqrt((this.getFamilyCount()) / GROUP_DENSITY / Math.PI + 0.5);
    }

    public int getFamilyCount() {
	int sizeCounter = 1;
	if (hasChildren()) {
	    for (Char child : children) {
		sizeCounter += child.getFamilyCount();
	    }
	}
	return sizeCounter;
    }

    public int getOptFamilyCount() {
	int sizeCounter = 0;
	for (int i = 0; i <= familyDegree; i++) {
	    sizeCounter += Math.pow(optChildrenCount, i);
	}
	return sizeCounter;
    }

    public boolean hasParent() {
	if (this.parent != null)
	    return true;
	else
	    return false;
    }

    public boolean hasChildren() {
	if (this.children.isEmpty()) {
	    return false;
	} else {
	    return true;
	}
    }

    public void addToChild(Char c) {
	this.children.add(c);
    }

    public String getName() {
	return this.name;
    }
    
    public char getSymbol() {
	return symbol;
    }
    
    public Color getColor() {
	return color;
    }

    public Cell getCell() {
	return this.cell;
    }

    public Cell getOrigCell() {
	return this.origCell;
    }

    public int getMovingSpeed() {
	return cell.getGroundProto().getSpeed() * this.speed / 100;
    }
    
    public int getActionSpeed() {
	return actionSpeed;
    }
    
    public void setSpeed(int newSpeed) {
	speed = newSpeed;
    }

    public double getMovingPr() {
	return this.movingPr;
    }

    public int getMovingR() {
	return this.movingR;
    }

    public double getHp() {
	return this.hp;
    }

    public double getMaxHp() {
	return this.maxHp;
    }

    public double getMp() {
	return this.mp;
    }

    public double getMaxMp() {
	return this.maxMp;
    }

    public boolean isHostile() {
	return hostile;
    }

    public double getTurnPool() {
	return turnPool;
    }

    public int getLevel() {
	return level;
    }

    public Map getMap() {
	return map;
    }
    
    public void setMap(Map target) {
	map = target;
    }

    public int getMinDmg() {
	return minDmg;
    }

    public int getMaxDmg() {
	return maxDmg;
    }
    
    public boolean isRideable() {
	return rideable;
    }

    public String getTalk() {
	return talk;
    }
    
    public void setTalk(String newTalk) {
	talk = newTalk;
    }
    
    public void ignite() {
	// TODO
    }
}
