package chars;

import items.Item;
import items.ItemPrototype;
import items.OwnerType;

import java.util.ArrayList;

import maps.Cell;
import maps.GroundType;
import util.Dir;
import util.Util;

public class NPC extends Char {

    private OwnerType type;
    private GroundType groundType;

    public NPC(String name, Cell c, int familyDegree, int optChildrenCount) {
	super(name, c);
	Util.log("Creating NPC: " + name + " at (" + origCell.getX() + "," + origCell.getY() + ")");
	NPCPrototype proto = NPCPrototype.getPrototype(name);
	this.name = proto.getName();
	this.type = proto.getType();
	symbol = proto.getSymbol();
	color = proto.getColor();
	this.maxHp = proto.getMaxHp();
	this.hp = this.maxHp;
	this.speed = proto.getSpeed();
	this.movingPr = proto.getMovingPr();
	this.movingR = proto.getMovingR();
	this.sight = proto.getSight();
	this.groundType = proto.getGroundType();
	this.hostile = proto.isHostile();
	this.minDmg = proto.getMinDmg();
	this.maxDmg = proto.getMaxDmg();
	this.familyDegree = familyDegree;
	for (int i = 0; i < familyDegree; i++) {		// TODO: add the 'boss' effect
	   // this.color = ColorData.moreGrey(this.color);
	}
	this.optChildrenCount = optChildrenCount;
	this.level = ((int) this.maxHp + (int) this.maxMp + this.minDmg + this.maxDmg) / 50;
	this.rideable = proto.isRideable();
	this.talk = proto.getTalk();
	this.hasAffair = false; // TODO: Char-Player-NPC-erottelu
    }

    public void doAction() { // TODO: EVERYTHING
	actionSpeed = getMovingSpeed();
	Player player = map.getGameModel().getPlayer();
	int familyCount = getFamilyCount();
	double maxParentDist = 0;
	double maxChildDist = getGroupRadius() * CHILD_DISTANCE_FACTOR;
	if (hasParent()) {
	    maxParentDist = parent.getGroupRadius();
	}

	if (familyDegree > 0 && familyCount < getOptFamilyCount()) {
	    generateChildren();
	}
	if (player.getRidingChar() == this) {
	    move(player.getCell());
	} else if (player.getMap() == map) {
	    double playerDist = distanceTo(player);
	    if (hasParent() && maxParentDist < distanceTo(parent) && maxParentDist < playerDist) {
		moveTowards(parent);
	    } else if (hasChildren() && hostile && playerDist < 3 && (3 * (familyCount - 1) > 2 * (getOptFamilyCount() - 1))) {
		if (isNextTo(player)) {
		    moveAwayFrom(player);
		} else if (hasChildren() && maxChildDist < playerDist && hasChildTooFar(maxChildDist)) {
		    moveTowards(firstChildTooFar(maxChildDist));
		} else {
		    actionSpeed = speed;
		}
	    } else if (hostile && isNextTo(player)) {
		attack(player);
		actionSpeed = speed;
	    } else if (hostile && sees(player)) {
		moveTowards(player);
	    } else if (distanceTo(origCell) > movingR) {
		moveTowards(origCell);
	    } else {
		stroll();
	    }

	} else {
	    if (hasParent() && maxParentDist < distanceTo(parent)) {
		moveTowards(parent);
	    } else if (hasChildren() && hasChildTooFar(maxChildDist)) {
		moveTowards(firstChildTooFar(maxChildDist));
	    } else {
		stroll();
	    }
	}
    }

    public boolean tryToMove(Dir dir) {
	int newX = cell.getX() + dir.getDeltaX();
	int newY = cell.getY() + dir.getDeltaY();
	if (newX >= map.getWidth() || newX < 0 || newY >= map.getHeight() || newY < 0) {
	    return false;
	}

	Cell target = map.getCells()[newX][newY];

	if ((Util.getDistance(newX, newY, origCell.getX(), origCell.getY()) <= movingR || distanceTo(origCell) > movingR) && (target.isFree() && target.getGroundProto().getType() == groundType || (target.getX() == cell.getX() && target.getY() == cell.getY()))) {
	    move(target);
	    return true;
	} else
	    return false;
    }

    protected void stroll() {
	if (this.movingPr > Math.random()) {
	    tryToMove(Dir.randomDir());
	}
    }

    protected void moveTowards(Cell target) {
	moveRelatively(target, true);
    }

    protected void moveTowards(Char c) {
	this.moveRelatively(c.getCell(), true);
    }

    protected void moveAwayFrom(Char c) {
	this.moveRelatively(c.getCell(), false);
    }

    private void moveRelatively(Cell target, boolean towards) {
	int x = cell.getX();
	int y = cell.getY();
	int tX = target.getX();
	int tY = target.getY();
	int dx = 0;
	int dy = 0;
	if (x < tX)
	    dx = 1;
	else if (x > tX)
	    dx = -1;
	if (y < tY)
	    dy = 1;
	else if (y > tY)
	    dy = -1;
	if (!towards) {
	    dx = (-1) * dx;
	    dy = (-1) * dy;
	}

	if (tryToMove(Dir.getDir(dx, dy))) {
	    return;
	}
	if (Math.abs(x - tX) > Math.abs(y - tY)) {
	    tryToMove(Dir.getDir(dx, 0));
	} else {
	    tryToMove(Dir.getDir(0, dy));
	}

	if (!towards && dy == 0) {
	    if (Math.random() < 0.5)
		if (tryToMove(Dir.getDir(dx, -1)) || tryToMove(Dir.getDir(dx, 1)))
		    return;
		else if (tryToMove(Dir.getDir(dx, 1)) || tryToMove(Dir.getDir(dx, -1)))
		    return;
	}
	if (!towards && dx == 0) {
	    if (Math.random() < 0.5)
		if (tryToMove(Dir.getDir(-1, dy)) || tryToMove(Dir.getDir(1, dy)))
		    return;
		else if (tryToMove(Dir.getDir(1, dy)) || tryToMove(Dir.getDir(-1, dy)))
		    return;
	}
    }

    private void generateChildren() {
	int childsToGenerate = optChildrenCount - children.size();
	if (CHILD_GENERATION_FACTOR * childsToGenerate / optChildrenCount > Math.random()) {
	    ArrayList<Cell> adjCells = cell.getAdjacentCells();
	    ArrayList<Cell> optCells = new ArrayList<Cell>();
	    for (Cell c : adjCells) {
		if (c.isFree() && c.getGroundProto().getType() == groundType)
		    optCells.add(c);
	    }
	    if (!optCells.isEmpty()) {
		Cell randomCell = optCells.get(Util.randInt(0, (optCells.size() - 1)));
		Char child = new NPC(name, randomCell, (familyDegree - 1), optChildrenCount);
		map.addChar(child);
		child.setParent(this);
		addToChild(child);
	    }
	}
    }

    private boolean hasChildTooFar(double maxChildDistance) {
	for (Char child : children) {
	    if (distanceTo(child) > maxChildDistance)
		return true;
	}
	return false;
    }

    private Char firstChildTooFar(double maxChildDistance) {
	for (Char child : children) {
	    if (distanceTo(child) > maxChildDistance)
		return child;
	}
	return null;
    }

    public void die() {
	dropItems();
	super.die();
    }

    private void dropItems() {
	if (Util.randInt(1, 2) == 2) {
	    dropItem(ItemPrototype.generateDroppedItem(this));
	}
    }

    private void dropItem(Item item) {
	item.setCell(this.getCell());
	this.getCell().getItems().add(item);
    }

    public boolean seesPlayer() {
	Cell playerCell = map.getGameModel().getPlayer().getCell();
	return (Util.getDistance(getCell().getX(), getCell().getY(), playerCell.getX(), playerCell.getY()) < sight);
    }

    public OwnerType getType() {
	return type;
    }

    public GroundType getGroundType() {
	return groundType;
    }
    
    public double getUsedTime() {
	return 100.0 / actionSpeed;
    }
}
