package chars;

import items.Item;

import java.util.ArrayList;

import maps.Cell;
import maps.GroundType;
import maps.Map;
import models.GameModel;
import skills.Skill;
import spells.Invisibility;
import spells.Spell;
import spells.WaterBolt;
import ui.Jorma;
import util.ColorData;
import util.Dir;
import util.Util;
import events.ContactEvent;
import events.Event;
import events.ItemEvent;
import events.MiscEvent;
import events.MoveEvent;

public class Player extends Char {

    private static final char PLAYER_SYMBOL = '@';
    private static final int PLAYER_GENERAL_SPEED = 100;
    private static final int PLAYER_HITTING_SPEED = 80;
    private static final int PLAYER_RIDING_SPEED = 120;

    private static final int PLAYER_MIN_DMG = 5;
    private static final int PLAYER_MAX_DMG = 10;
    private static final int PLAYER_MAX_HP = 1000;
    private static final int PLAYER_MAX_MP = 60;
    private static final int PLAYER_SIGHT = 8;

    private static final int WALKING = 1;
    private static final int RIDING = 2;
    private static final int SWIMMING = 3;
    private static final int LEVITATING = 4;
    private static final int GHOSTING = 5;

    private Race race = Race.HIGH_ELF;
    private int exp = 0;
    private int expNeeded = 10;
    private ArrayList<Item> items;
    private ArrayList<Spell> spells;
    private double itemsWeight;
    private int defense;
    private int sorcery;
    private int swimming;
    private Char ridingChar;
    private int generalSpeed;
    private int hittingSpeed;
    private int ridingSpeed;
    private int movingType;

    public Player(String name, Cell cell) {
	super(name, cell);
	symbol = PLAYER_SYMBOL;
	color = ColorData.getColor("yellow", 3);
	//symbol = PLAYER_SYMBOL;
	generalSpeed = PLAYER_GENERAL_SPEED;
	hittingSpeed = PLAYER_HITTING_SPEED;
	ridingSpeed = PLAYER_RIDING_SPEED;
	speed = generalSpeed;
	minDmg = PLAYER_MIN_DMG;
	maxDmg = PLAYER_MAX_DMG;
	this.maxHp = PLAYER_MAX_HP;
	this.hp = maxHp;
	this.maxMp = PLAYER_MAX_MP;
	this.mp = maxMp;
	this.level = 1;
	this.sight = PLAYER_SIGHT;
	//color = ColorData.getColor("yellow", 3);
	setMap(map);
	ridingChar = null;
	rideable = false;
	movingType = WALKING;

	items = new ArrayList<Item>();
	items.add(new Item("small club", this));
	items.add(new Item("long club", this));
	items.add(new Item("mushroom", this));
	items.add(new Item("mug", this));
	items.add(new Item("katana", this));
	items.add(new Item("small club", this));
	items.add(new Item("long club", this));
	items.add(new Item("small axe", this));

	spells = new ArrayList<Spell>();
	spells.add(new WaterBolt(this));
	spells.add(new Invisibility(this));

	updateItemsWeight();
    }

    public void attack(Char target) {
	int dmg = calcDmg(target);
	target.decreaseHp(dmg);
	String msg = "You hit " + dmg + " to " + Util.capitalizeFirstLetter(target.getName()) + ".";
	addEvent(new ContactEvent(hittingSpeed, msg, Event.ATTACK, target));
    }

    public void attack(Cell target) {
	int dmg = maxDmg;
	target.decreaseHp(dmg);
	String msg = "You hit " + dmg + " to " + target.getGroundProto().getName() + ".";
	addEvent(new MiscEvent(hittingSpeed, msg, Event.ATTACK)); // TODO: does attack() event need oldCell really?
    }

    private int calcDmg(Char target) {
	return Util.randInt(minDmg, maxDmg);
    }

    protected void die() {
	map.getGameModel().getWindow().getJorma().setCurrentModel(Jorma.MODEL_HIGHSCORES);
    }

    public void updateItemsWeight() {
	itemsWeight = 0;
	for (Item i : items) {
	    itemsWeight += i.getWeight();
	}
    }

    public double getItemsWeight() {
	return itemsWeight;
    }

    public void setMap(Map map) {
	if (this.map != null) {
	    this.map.getGameModel().setPlayer(null);
	}
	this.map = map;
	map.getGameModel().setPlayer(this);
    }

    public void recovery() {
	increaseHp(maxHp * (RECOVERY_FACTOR + 0.002 * Skill.RECOVERY.getLevel()));
	increaseMp(maxMp * (RECOVERY_FACTOR + 0.002 * Skill.MEDITATING.getLevel()));
    }

    public void move(Cell target) {
	Cell oldCell = cell;
	super.move(target);
	int speed = 0;
	if (movingType == LEVITATING || movingType == GHOSTING)
	    addEvent(new MoveEvent(generalSpeed, Event.MOVE, oldCell, target));
	else {
	    switch (movingType) {
	    case RIDING:
		speed = ridingSpeed;
		break;
	    default:
		speed = generalSpeed;
		break;
	    }
	    addEvent(new MoveEvent(target.getGroundProto().getSpeed() * speed / 100, Event.MOVE, oldCell, target));
	}
    }

    public void doAction(Dir dir) {
	map.getGameModel().clearOutput();
	int x = getCell().getX();
	int y = getCell().getY();
	int newX = x + dir.getDeltaX();
	int newY = y + dir.getDeltaY();

	if (newX < map.getWidth() && newX >= 0 && newY < map.getHeight() && newY >= 0) {
	    Cell target = map.getCells()[newX][newY];
	    if (!target.isFree() && target != cell) {
		Char c = target.getChar();
		if (c.isHostile()) {
		    attack(c);
		} else {
		    talkTo(c);
		}
	    } else {
		GroundType targetType = target.getGroundProto().getType();
		switch (movingType) {
		case WALKING:
		    if (targetType.isWalkable())
			move(target);
		    else if (targetType.isSwimmable()) {
			if (Skill.SWIMMING.isLearned()) {
			    movingType = SWIMMING;
			    move(target); // TODO: CHANGE MOVING STYLE needed anyway? Message?
			    addEvent(new MiscEvent(0, "Swimming.", Event.CHANGE_MOVING_STYLE)); // TODO: replace 100 with a moveTransition speed?
			} else {
			    addEvent(new MiscEvent(generalSpeed, "Can't swim.", Event.FAILED_ACTION));
			}
		    } else {
			addEvent(new MiscEvent(generalSpeed, "Ouch! Can't walk through " + target.getGroundProto().getName() + ".", Event.FAILED_ACTION));
		    }
		    break;
		case RIDING:
		    if (targetType.isRideable())
			move(target);
		    else {
			addEvent(new MiscEvent(generalSpeed, "Can't ride that way.", Event.FAILED_ACTION));
		    }
		    break;
		case SWIMMING:
		    if (targetType.isSwimmable() && hasGroundNear(target))
			move(target);
		    else if (targetType.isWalkable()) {
			movingType = WALKING;
			move(target);
			addEvent(new MiscEvent(0, "Walking.", Event.CHANGE_MOVING_STYLE));
		    } else {
			addEvent(new MiscEvent(generalSpeed, "Can't get there.", Event.FAILED_ACTION));
		    }
		    break;
		case LEVITATING:
		    if (targetType.isLevitable()) {
			move(target);
			if (mp > 5)
			    mp -= 5;
			else {
			    movingType = WALKING;
			    addEvent(new MiscEvent(0, "Back to use your feet, no mp anymore.", Event.CHANGE_MOVING_STYLE));
			}
		    } else {
			addEvent(new MiscEvent(generalSpeed, "Can't levitate there.", Event.FAILED_ACTION));
		    }
		    break;
		case GHOSTING:
		    if (!targetType.isSwimmable() && Skill.GHOSTING.isLearned())
			move(target);
		    else {
			addEvent(new MiscEvent(0, "Something wrong there...", Event.MESSAGE));
		    }
		    movingType = WALKING;
		    break;
		default:
		    addEvent(new MiscEvent(0, "Isn't moving by any known method...", Event.MESSAGE));
		}
	    }
	} else {
	    addEvent(new MiscEvent(generalSpeed, "Hmm... I don't want to die in space.", Event.FAILED_ACTION));
	}
    }

    public void tryToStartRiding(Dir dir) {
	map.getGameModel().clearOutput();
	Cell dirCell = cell.getAdjacentCell(dir);
	Char ridingChar = dirCell.getChar();
	if (ridingChar != null && ridingChar.isRideable()) {
	    setRidingChar(ridingChar);
	    setMovingType(RIDING);
	    move(dirCell);
	    addEvent(new ContactEvent(0, "Riding.", Event.START_RIDING, ridingChar));
	} else {
	    addEvent(new MiscEvent(generalSpeed, "Can't ride that way.", Event.FAILED_ACTION));
	}
    }

    public void tryToStopRiding(Dir dir) {
	map.getGameModel().clearOutput();
	Cell dirCell = cell.getAdjacentCell(dir);
	Cell cellToLeave = cell;
	Char charToLeave = ridingChar;
	if (dirCell.isFree() && dirCell.getGroundProto().getType().isWalkable()) {
	    setRidingChar(null);
	    move(dirCell);
	    setMovingType(WALKING);
	    charToLeave.move(cellToLeave);
	    charToLeave.setOrigCell(cellToLeave);
	    addEvent(new ContactEvent(0, "Walking.", Event.STOP_RIDING, charToLeave));
	} else {
	    addEvent(new MiscEvent(generalSpeed, "Can't come down there.", Event.FAILED_ACTION));
	}
    }

    public void tryToAttack(Dir dir) {
	Char c = cell.getAdjacentCell(dir).getChar();
	if (c != null)
	    attack(c);
	else {
	    GroundType dirType = cell.getAdjacentCell(dir).getGroundProto().getType();
	    if (!dirType.isWalkable() && !dirType.isSwimmable()) {
		attack(cell.getAdjacentCell(dir));
	    } else {
		addEvent(new MiscEvent(generalSpeed, "Isn't appropriate to hit that.", Event.FAILED_ACTION));
	    }
	}
    }

    public void tryToTalk(Dir dir) {
	Char c = cell.getAdjacentCell(dir).getChar();
	if (c != null)
	    talkTo(c);
	else {
	    addEvent(new MiscEvent(generalSpeed, "There's nobody to talk with.", Event.FAILED_ACTION));
	}
    }

    private boolean hasGroundNear(Cell target) {
	int x = target.getX();
	int y = target.getY();
	for (int i = -swimming; i <= swimming; i++) {
	    for (int j = -swimming; j <= swimming; j++) {
		if (x + i >= 0 && y + j >= 0 && x + i < map.getWidth() && y + j < map.getHeight() && Util.getDistance(x, y, x + i, y + j) <= swimming && map.getCells()[x + i][y + j].getGroundProto().getType().isWalkable()) {
		    return true;
		}

	    }
	}
	return false;
    }

    private void talkTo(Char c) {
	String talk = c.getTalk();
	String msg;
	if (talk == null) {
	    msg = "You can't talk with " + c.getName() + ".";
	} else {
	    msg = Util.capitalizeFirstLetter(c.getName()) + ": \"" + c.getTalk() + "\"";
	}
	addEvent(new ContactEvent(generalSpeed, msg, Event.TALK, c));
    }

    public int getMovingType() {
	return movingType;
    }

    private void setMovingType(int type) {
	movingType = type;
    }

    public void update() {
	recovery();
	updateSight();
	updateSkills();
    }

    private void updateSight() {
	int x = getCell().getX();
	int y = getCell().getY();
	for (int i = -sight - 1; i < sight + 1; i++) {
	    for (int j = -sight - 1; j < sight + 1; j++) {
		Cell c;
		if (x + i >= 0 && y + j >= 0 && x + i < map.getWidth() && y + j < map.getHeight() && (c = map.getCells()[x + i][y + j]).getVisibility() == Cell.CURRENTLY_SEEN) {
		    c.setVisibility(Cell.KNOWN);
		}
	    }
	}

	updateSightRange();

	for (double angle = 0.0; angle < 2 * Math.PI; angle += 1 / (double) sight / 8) {
	    boolean wall = false;
	    for (double r = 0; r < sight; r++) {
		if (!wall) {
		    int traversalX = (int) Math.round(r * Math.cos(angle));
		    int traversalY = (int) Math.round(r * Math.sin(angle));
		    if (x + traversalX >= 0 && y + traversalY >= 0 && x + traversalX < map.getWidth() && y + traversalY < map.getHeight()) {
			Cell c = map.getCells()[x + traversalX][y + traversalY];
			c.setVisibility(Cell.CURRENTLY_SEEN);
			if (c.getGroundProto().getType().isBlocksSight()) {
			    wall = true;
			}
		    } else {
			wall = true;
		    }
		}
	    }
	}
    }

    private void updateSightRange() {
	int timeOfDay = map.getGameModel().getTimeOfDay();
	switch (timeOfDay) {
	case GameModel.DAY:
	    sight = PLAYER_SIGHT;
	    break;
	case GameModel.MORNING:
	case GameModel.EVENING:
	    sight = PLAYER_SIGHT * 3 / 4;
	    break;
	case GameModel.NIGHT:
	    sight = PLAYER_SIGHT * 3 / 8;
	    break;
	default:
	    sight = PLAYER_SIGHT;
	    break;
	}
	Skill.STRENGTH.learn();
    }

    public void updateSkills() {
	swimming = Skill.SWIMMING.getLevel();
    }

    public Race getRace() {
	return race;
    }

    public void addExp(int amount) {
	exp += amount;

	if (exp > expNeeded) {
	    levelUp();
	}
    }

    private void levelUp() {
	level++;
	exp = 0;
	expNeeded = (int) (expNeeded * 1.3);
	maxHp = (int) (maxHp * 1.1);
	maxMp = (int) (maxMp * 1.1);
    }

    public int getExp() {
	return exp;
    }

    public int getExpNeeded() {
	return expNeeded;
    }

    public ArrayList<Item> getItems() {
	return items;
    }

    public void pickUp(Item item) {
	cell.getItems().remove(item);
	items.add(item);
	updateItemsWeight();
	item.setPlayer(this);
	item.setCell(null);
	String msg = "You picked up " + item.getName() + ".";
	addEvent(new ItemEvent(generalSpeed, msg, Event.PICK_UP, item));
    }

    public void setOnFire() {
	cell.tryToSetOnFire(true);
    }

    public ArrayList<Spell> getSpells() {
	return spells;
    }

    public int getDefense() {
	return defense;
    }

    public int getSorcery() {
	return sorcery;
    }

    public Char getRidingChar() {
	return ridingChar;
    }

    public boolean isRiding() {
	if (ridingChar != null)
	    return true;
	else
	    return false;
    }

    public void setRidingChar(Char c) {
	ridingChar = c;
    }

    public void levitate() {
	if (movingType == LEVITATING)
	    movingType = WALKING;
	else if (mp > 5)
	    movingType = LEVITATING;
	else {
	    addEvent(new MiscEvent(generalSpeed, "Can't levitate just now.", Event.FAILED_ACTION));
	}
    }

    public void ghost() {
	if (movingType == GHOSTING)
	    movingType = WALKING;
	else if (Skill.GHOSTING.isLearned())
	    movingType = GHOSTING;
	else {
	    addEvent(new MiscEvent(generalSpeed, "Can't ghost.", Event.FAILED_ACTION));
	}
    }

    private void addEvent(Event e) {
	map.getGameModel().addEvent(e);
    }
}
