package controllers;

import items.Item;

import java.awt.event.KeyEvent;
import java.util.ArrayList;

import maps.Cell;
import maps.Link;
import maps.Map;
import models.GameModel;
import models.InventoryModel;
import models.SkillsModel;
import models.SpellsModel;
import skills.Skill;
import spells.Spell;
import ui.Jorma;
import ui.Window;
import util.Dir;
import chars.Player;
import events.MiscEvent;

public class GameController extends Controller {

    private static final int ATTACK = KeyEvent.VK_F;
    private static final int CANCEL = KeyEvent.VK_ESCAPE;
    private static final int CAST = KeyEvent.VK_C;
    private static final int ENTER = KeyEvent.VK_ENTER;
    private static final int GHOST = KeyEvent.VK_W;
    private static final int HELP = KeyEvent.VK_F10;
    private static final int INVENTORY = KeyEvent.VK_I;
    private static final int LEVITATE = KeyEvent.VK_Q;
    private static final int PICK_UP = KeyEvent.VK_COMMA;
    private static final int RIDE = KeyEvent.VK_R;
    private static final int SET_ON_FIRE = KeyEvent.VK_S;
    private static final int SHOW_OUTPUT = KeyEvent.VK_SPACE;
    private static final int SKILLS = KeyEvent.VK_Z;
    private static final int TALK = KeyEvent.VK_A;
    private static final int WAIT = KeyEvent.VK_L;

    private static final int NOT_WAITING = 0;

    private int directedAct = NOT_WAITING;

    public GameController(GameModel model) {
	super(model);
    }

    public void keyPressed(KeyEvent e) {
	GameModel m = (GameModel) model;
	boolean outputWaiting = m.isOutputWaiting();
	Spell spellBeingCasted = m.getSpellBeingCasted();
	Player player = m.getPlayer();
	Window window = m.getWindow();

	int keyCode = e.getKeyCode();
	if (outputWaiting) {
	    if (keyCode == SHOW_OUTPUT) {
		m.setVisibleOutput("");
		m.updateVisibleOutput();
	    }
	} else if (spellBeingCasted != null) {
	    Dir dir = Dir.keyToDir(e);
	    if (dir != null && dir != Dir.NONE) {
		spellBeingCasted.cast(dir);
		m.setSpellBeingCasted(null);
		m.newTurn();
	    }
	} else if (directedAct != NOT_WAITING) {
	    Dir dir = Dir.keyToDir(e);
	    switch (directedAct) {
	    case RIDE:
		if (dir != null) {
		    if (player.isRiding())
			player.tryToStopRiding(dir);
		    else
			player.tryToStartRiding(dir);
		    m.newTurn();
		} else if (keyCode == CANCEL) {
		    m.clearOutput();
		    m.addMessage("Forgot it...");
		}
		break;
	    case ATTACK:
		m.clearOutput();
		if (dir != null) {
		    player.tryToAttack(dir);
		    m.newTurn();
		} else if (keyCode == CANCEL) {
		    m.clearOutput();
		    m.addMessage("Forgot it...");
		}
		break;
	    case TALK:
		m.clearOutput();
		if (dir != null) {
		    player.tryToTalk(dir);
		    m.newTurn();
		} else if (keyCode == CANCEL) {
		    m.addMessage("Forgot it...");
		}
		break;
	    }
	    directedAct = NOT_WAITING;
	} else {
	    switch (keyCode) {
	    case HELP:
		window.getJorma().setCurrentModel(Jorma.MODEL_HELP);
		break;
	    case INVENTORY:
		m.clearOutput();
		InventoryModel inventoryModel = (InventoryModel) Jorma.MODEL_INVENTORY;
		inventoryModel.updateList();
		window.getJorma().setCurrentModel(inventoryModel);
		break;
	    case CAST:
		m.clearOutput();
		SpellsModel spellsModel = (SpellsModel) Jorma.MODEL_SPELLS;
		spellsModel.updateList(player.getSpells());
		window.getJorma().setCurrentModel(spellsModel);
		break;
	    case WAIT:
		m.waitTenTurns();
		break;
	    case RIDE:
		m.clearOutput();
		directedAct = RIDE;
		m.addMessage("Give riding action or press Esc.");
		break;
	    case ATTACK:
		m.clearOutput();
		directedAct = ATTACK;
		m.addMessage("Give attack direction or press Esc.");
		break;
	    case TALK:
		m.clearOutput();
		directedAct = TALK;
		m.addMessage("Give talk direction or press Esc.");
		break;
	    case PICK_UP:
		ArrayList<Item> itemsLying = player.getCell().getItems();
		if (itemsLying.size() > 0) {
		    player.pickUp(itemsLying.get(itemsLying.size() - 1));
		    m.newTurn();
		}
		break;
	    case SET_ON_FIRE:
		player.setOnFire();
		break;
	    case LEVITATE:
		player.levitate();
		m.newTurn();
		break;
	    case GHOST: // TODO: make it a spell, no dummy buttoning around my boy... Means the VK_S n' VK_Q too
		player.ghost();
		m.newTurn();
		break;
	    case ENTER:
		Link link = player.getCell().getLink();
		if (link != null) {
		    Map targetMap = m.getMaps().get(link.getTargetMapName());
		    Cell targetCell = targetMap.getCells()[link.getTargetX()][link.getTargetY()];
		    m.clearOutput();
		    if (targetCell.isFree()) {
			m.setCurrentMap(targetMap);
			player.setMap(targetMap);
			player.move(targetCell);
			m.addMessage("You arrived to " + m.getCurrentMap().getName() + ".");
		    } else {
			targetCell.getChar().tryToMove(Dir.randomDir());
			m.addEvent(new MiscEvent(100, "Something was on your way there. You tried to get it moving.", MiscEvent.FAILED_ACTION));
			// TODO: speed -> generalSpeed == move this to m
		    }
		    m.updateVisibleOutput();
		    m.newTurn();
		}
		break;
	    case SKILLS:
		SkillsModel skillsModel = (SkillsModel) Jorma.MODEL_SKILLS;
		skillsModel.updateList(Skill.getLearnedSkills());
		window.getJorma().setCurrentModel(skillsModel);
	    default:
		Dir dir = Dir.keyToDir(e);
		if (dir != null) {
		    player.doAction(dir);
		    m.updateVisibleOutput();
		    m.newTurn();
		}
	    }
	}
    }
}
