package maps;

import items.Item;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.LinkedList;

import lists.BufferedList;
import models.GameModel;
import spells.Spell;
import util.Dir;
import util.Util;
import chars.Char;
import chars.NPC;

public class Map {

    private GameModel gameModel;
    private Cell[][] cells;
    private int width, height;
    private String name;
    private String filename;
    private BufferedList<Char> chars = new BufferedList<Char>();
    private BufferedList<Spell> affectingSpells = new BufferedList<Spell>();
    private BufferedList<Cell> burningCells = new BufferedList<Cell>();

    public Map(GameModel gameModel, String name, String filename) {
	this.gameModel = gameModel;
	this.name = name;
	this.filename = filename;

	load();
    }

    public void load() {
	try {
	    BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "UTF-8"));

	    // Grounds
	    String[] dimensions = reader.readLine().split(" ");
	    width = Integer.parseInt(dimensions[0]);
	    height = Integer.parseInt(dimensions[1]);

	    cells = new Cell[width][height];
	    String mapRow;
	    for (int i = 0; i < height; i++) {
		mapRow = reader.readLine();

		for (int j = 0; j < width; j++) {
		    cells[j][i] = new Cell(this, mapRow.charAt(j), j, i);
		}
	    }

	    int numOfMonsters = Integer.parseInt(reader.readLine());

	    // Monsters
	    for (int i = 0; i < numOfMonsters; i++) {
		mapRow = reader.readLine();
		String[] parts = mapRow.split(";");
		chars.add(new NPC(parts[0], cells[Integer.parseInt(parts[1])][Integer.parseInt(parts[2])], Integer.parseInt(parts[3]), Integer.parseInt(parts[4])));
	    }

	    int numOfItems = Integer.parseInt(reader.readLine());

	    // Items

	    for (int i = 0; i < numOfItems; i++) {
		mapRow = reader.readLine();
		String[] parts = mapRow.split(";");
		cells[Integer.parseInt(parts[1])][Integer.parseInt(parts[2])].addItem(new Item(parts[0]));
	    }

	    int numOfStairs = Integer.parseInt(reader.readLine());

	    // Stairs

	    for (int i = 0; i < numOfStairs; i++) {
		mapRow = reader.readLine();
		String[] parts = mapRow.split(";");
		cells[Integer.parseInt(parts[0])][Integer.parseInt(parts[1])].setLink(new Link(parts[2], Integer.parseInt(parts[3]), Integer.parseInt(parts[4])));
	    }

	    reader.close();
	} catch (Exception e) {
	    Util.logError("Couldn't read map data: " + name);
	}
    }
    
    public void update(double timeLeap) {
	for (Char c : chars) {
	    c.increaseTurnPool(timeLeap);
	    while (c.getTurnPool() > 0) {
		c.doAction();
		c.decreaseTurnPool(100.0 / c.getActionSpeed());
		c.recovery();
	    }
	}
	chars.update();
	updateAffectingSpells();
	updateBurningCells();
    }

    private void updateAffectingSpells() {
	for (Spell s : affectingSpells) {
	    s.decreaseDuration();
	    if (s.getDuration() == 0)
		affectingSpells.remove(s);
	}
	affectingSpells.update();
    }
    
    public void addAffectingSpell(Spell spell) {
	affectingSpells.add(spell);
    }
    
    private void updateBurningCells() {
	for (Cell c : burningCells) {
	    c.burn();
	    if (!c.burning())
		burningCells.remove(c);
	}
	burningCells.update();
    }
    
    public Dir nextDirInPath(Cell start, Cell end) {
	if (start == end) {
	    return Dir.NONE;
	}
	
	for (int i = 0; i < cells.length; i++) {
	    for (int j = 0; j < cells[0].length; j++) {
		cells[i][j].setPrevDir(null);
		cells[i][j].setVisited(false);
	    }
	}

	LinkedList<Cell> queue = new LinkedList<Cell>();
	queue.push(start);

	Cell c;
	do {
	    c = queue.pop();

	    if (c == end) {
		break;
	    }

	    for (Dir dir : Dir.DIRS) {
		if (dir != Dir.NONE) {
		    Cell adj = c.getAdjacentCell(dir);
		    if (adj != null && !adj.isVisited() && adj.getGroundProto().getType().isWalkable() && (adj.isFree() || adj == end)) {
			adj.setVisited(true);
			adj.setPrevDir(dir);
			queue.addLast(adj);
		    }
		}
	    }
	} while (queue.size() > 0);

	if (c != end) {
	    return Dir.NONE;
	}

	Cell adj;
	while ((adj = c.getAdjacentCell(c.getPrevDir().getOppositeDir())) != start) {
	    c = adj;
	}

	return c.getPrevDir();
    }
    
    public void addChar(Char c) {
	chars.add(c);
    }

    public Cell getCellAt(int x, int y) {
	return cells[x][y];
    }

    public Cell[][] getCells() {
	return cells;
    }

    public GameModel getGameModel() {
	return gameModel;
    }

    public BufferedList<Char> getChars() {
	return chars;
    }

    public int getWidth() {
	return width;
    }

    public int getHeight() {
	return height;
    }

    public BufferedList<Spell> getAffectingSpells() {
	return affectingSpells;
    }
    
    public void toBurningList(Cell cell) {
	burningCells.add(cell);
    }
    
    public String getName() {
	return name;
    }

    public double getTimeLeap() {
	return gameModel.getTimeLeap();
    }
}
