package models;

import items.Item;

import java.util.ArrayList;

import lists.BufferedList;
import maps.Cell;
import maps.Map;
import maps.MapList;
import quests.Quest;
import quests.RidingQuest;
import quests.TestQuest;
import spells.Spell;
import ui.Window;
import util.Util;
import views.GameView;
import views.View;
import chars.Char;
import chars.Player;
import controllers.GameController;
import events.Event;
import events.MiscEvent;

public class GameModel extends Model {

    public static final int OUTPUTBOX_HEIGHT = 4;
    public static final int WIDTH_FOR_TEXT = (int) (View.WIDTH * 1.24);
    public static final int DAY_LENGTH = 24000;
    public static final int NIGHT = 1;
    public static final int MORNING = 2;
    public static final int DAY = 3;
    public static final int EVENING = 4;

    public static final String PLAYER_NAME = "jokunen";
    public static final String FIRST_MAP_NAME = MapList.VILLAGE_MAP_NAME;

    public static final int PREVIOUS_MAP_TIMER_INIT = 20;

    private Map currentMap;
    private Map previousMap;
    private double previousMapTimer;
    private MapList maps;
    private Player player = null;
    private int turn;
    private double timeLeap;
    private double time;
    private int day;
    private ArrayList<String> hours;
    private BufferedList<Event> events;

    private String outputTemp = "";
    private String visibleOutput = "";
    private boolean outputWaiting = false;

    private Spell spellBeingCasted = null;

    public GameModel(Window window) {
	super(window);
	this.view = new GameView(this);
	this.controller = new GameController(this);
	maps = new MapList(this);
	currentMap = maps.get(FIRST_MAP_NAME);
	setCurrentMap(currentMap);
	previousMap = null;
	previousMapTimer = 0;
	player = new Player(PLAYER_NAME, currentMap.getCells()[8][8]);		// luodaan se playeri oikein
	turn = 0;
	timeLeap = 0;
	time = 16955; // Util.randInt(DAY_LENGTH / 5, DAY_LENGTH / 2);
	day = 1;
	hours = generateHourNames();
	events = new BufferedList<Event>();
	initQuests();
	updateVisibleOutput();
    }

    public void newTurn() {
	turn++;
	updateTime();
	currentMap.update(timeLeap);
	if (previousMapTimer > 0) {
	    previousMap.update(timeLeap);
	    previousMapTimer -= timeLeap;
	}
	player.update();
	for (Event e : events.getAddBuffer()) {
	    addMessage(e.getMessage());
	}
	events.update(); // TODO: optimize double looping
    }
    
    public void waitTenTurns() {
	for (int i = 0; i < 10; i++) {
	    boolean hostileNPCNear = false;
	    for (Char c : currentMap.getChars()) {
		if (c.isNextTo(player) && c.isHostile()) {
		    hostileNPCNear = true;
		}
	    }
	    if (!hostileNPCNear) {
		addEvent(new MiscEvent(100, "", Event.OTHER_EVENT));
		newTurn();
	    } else {
		i = 10;
	    }
	}
	clearOutput();
	updateVisibleOutput();
    }

    public MapList getMaps() {
	return maps;
    }

    private void initQuests() {
	Quest.QUESTS.add(new TestQuest(this));	// TODO: Why is it static final? How to use getter or so instead?
	Quest.QUESTS.add(new RidingQuest(this));
    }

    public Player getPlayer() {
	return player;
    }

    public void setPlayer(Player player) {
	this.player = player;
    }

    public void addMessage(String s) {
	if (s != null) {
	    outputTemp += s + " ";
	    updateVisibleOutput();
	}
    }

    public void updateVisibleOutput() {
	String output = null;
	if (!outputWaiting) {
	    Cell playerCell = player.getCell();
	    ArrayList<Item> items = playerCell.getItems();
	    output = Util.capitalizeFirstLetter(player.getCell().getGroundProto().getName()) + ". ";
	    if (!items.isEmpty()) {
		if (items.size() > 1)
		    output += "Several items lying. ";
		else
		    output += Util.capitalizeFirstLetter(items.get(0).getName()) + " lying. ";
	    }
	    output += outputTemp;
	} else
	    output = outputTemp;

	output = Util.splitToLines(output, WIDTH_FOR_TEXT);
	int maxLength = (OUTPUTBOX_HEIGHT) * WIDTH_FOR_TEXT - (int) (WIDTH_FOR_TEXT * 0.7);
	if (visibleOutput.length() < maxLength) {
	    if (output.length() > maxLength) {
		visibleOutput = output.substring(0, maxLength) + " ... PRESS SPACE TO READ MORE";
		outputTemp = output.substring(maxLength);
		outputWaiting = true;
	    } else {
		visibleOutput = output;
		outputWaiting = false;
	    }
	}
    }

    public void clearOutput() {
	outputTemp = "";
	visibleOutput = "";
    }

    public void setSpellBeingCasted(Spell spellBeingCasted) {
	this.spellBeingCasted = spellBeingCasted;
    }

    public Map getCurrentMap() {
	return currentMap;
    }

    public void setCurrentMap(Map targetMap) {
	previousMap = currentMap;
	currentMap = targetMap;
	previousMapTimer = PREVIOUS_MAP_TIMER_INIT;
	((GameView) view).initMapView(targetMap);
    }

    public String getVisibleOutput() {
	return visibleOutput;
    }

    public void setVisibleOutput(String visibleOutput) {
	this.visibleOutput = visibleOutput;
    }

    public boolean isOutputWaiting() {
	return outputWaiting;
    }

    public void setOutputWaiting(boolean outputWaiting) {
	this.outputWaiting = outputWaiting;
    }

    public Spell getSpellBeingCasted() {
	return spellBeingCasted;
    }

    public int getTurn() {
	return turn;
    }

    public double getTime() {
	return time;
    }

    public double getTimeLeap() {
	return timeLeap;
    }

    public int getDay() {
	return day;
    }

    private void updateTime() {
	timeLeap = 0;
	for (Event e : events.getAddBuffer()) {
	    if (e.getSpeed() != 0)
		timeLeap += 100 / e.getSpeed();
	}
	time += timeLeap;
	if (time >= DAY_LENGTH) {
	    day++;
	    time -= DAY_LENGTH;
	}
    }

    public int getTimeOfDay() {
	if (time < DAY_LENGTH * 1 / 6)
	    return NIGHT;
	else if (time < DAY_LENGTH * 1 / 4)
	    return MORNING;
	else if (time < DAY_LENGTH * 3 / 4)
	    return DAY;
	else if (time < DAY_LENGTH * 5 / 6)
	    return EVENING;
	else
	    return NIGHT;
    }

    public String getTimeOfDayWord() {
	int timeOfDay = getTimeOfDay();
	switch (timeOfDay) {
	case NIGHT:
	    return "night";
	case MORNING:
	    return "morning";
	case DAY:
	    return "day";
	case EVENING:
	    return "evening";
	default:
	    return "no any";
	}
    }

    private static ArrayList<String> generateHourNames() {
	ArrayList<String> hours = new ArrayList<String>();
	hours.add("first");
	hours.add("second");
	hours.add("third");
	hours.add("fourth");
	hours.add("fifth");
	hours.add("sixth");
	hours.add("seventh");
	hours.add("eighth");
	hours.add("ninth");
	hours.add("tenth");
	hours.add("eleventh");
	hours.add("twelfth");
	return hours;
    }

    private String getHour() {
	if (time > DAY_LENGTH * 1 / 4 && time < DAY_LENGTH * 3 / 4)
	    return hours.get((int) time / (DAY_LENGTH / 24) - 6);
	return null;
    }

    public String getHourOutput() {
	if (getHour() != null)
	    return "It's " + getHour() + " hour.";
	else
	    return "";
    }

    public BufferedList<Event> getEvents() {
	return events;
    }

    public void addEvent(Event e) {
	events.add(e);
	Quest.updateConditions();
    }
}
